// *************************************************
// camera defect : circles of confusion (coc, bokeh)
// *************************************************
// Puts objects in front of the camera, just a bit over the viewing plane,
// by tracing rays from the camera to evaluate the image color and focus at 
// the traced point, using the color found to texturize the bokeh objects, 
// and the focus amount to scale them.

// a plane to trace into
#declare test_box=plane{z,1.1}

// pigment for evaluating the input image
#declare p_test=
pigment{
  input_image_map
  translate <-.5,-.5,0>
  scale 1+chromatic_aberration*aberration_amount*2 
}

// pigment for evaluating the "focus map"
#declare p_coc=
pigment{
  image_map{
    #switch (focus_map_type)
      #case (0)
        hdr
        #break
      #case (1)
        jpeg 
        #break
      #case (2)
        png
        #break
      #case (3)
        tga
        #break
    #end
    focus_map 
    once
    map_type 0
  }
  translate <-.5,-.5,0>
  scale 1+aberration_amount*2
}

// create the bokeh object
#declare i_blades=0;
#declare coc_obj=
mesh{
  #while (i_blades<coc_blades)
    triangle{<0,0,0>,vrotate(<0,coc_radius,0>,<0,0,360*i_blades/coc_blades>),vrotate(<0,coc_radius,0>,<0,0,360*(i_blades+1)/coc_blades>)}
    #declare i_blades=i_blades+1;
  #end
  translate -.01*z
}

// to avoid excesive overlaping of the bokeh objects, we will use an extra trace test 
// over a union of spheres representing previously placed objects, so we first create
// something to add the test speheres later:
#declare avoid_overlapping=box{-.5,.5 translate 10*y}  //just something away from the tracing area
// in case of HDR, it will be useful to know wich was the value of the brightest spot,
// so we can adjust the trhesold value for the next render
#declare max_val=0;
// start placing bokeh objects
#declare i_dens=0;
#while (i_dens<coc_density)
  // random point to trace again the text box
  #declare coc_pt=<-.5+rand(r_coc),-.5+rand(r_coc),1>;
  #declare Norm=<0,0,0>;
  #declare Inter=trace(test_box,coc_pt,z,Norm);
  // evaluate the image color at this point
  #declare coc_int=eval_pigment(p_test,Inter);
  // and the amount of focus at this point
  #declare coc_depth=eval_pigment(p_coc,Inter);
  // find the value of the brightest spot on the image
  #if (coc_int.gray>max_val) 
    #declare max_val=coc_int.gray; 
  #end
  // place the bokeh objects on spots brighter than the desired threshold
  #if (coc_int.gray>coc_threshold)
    // trace against the placement of previous bokeh objects
    #declare Norm=<0,0,0>;
    #declare Inter=trace(avoid_overlapping,coc_pt,z,Norm);
    // and place the object if the location is "free"
    #if (Norm.gray=0)
      object{coc_obj 
        texture{
          pigment{
            spherical poly_wave coc_pw
            color_map{
              [0 rgbft coc_int+<0,0,0,0,1>]
              [1 rgbft coc_int+<0,0,0,0,(1-coc_strength)>]
            }
            scale coc_radius
          }
          finish{emission 1}
        }
        scale (pow(coc_depth.gray,coc_pw)+.0000001)
        rotate coc_angle*z
        translate coc_pt*<width_ratio,height_ratio,1>
        // the effect works best if brightests spots are placed behind 
        translate -coc_int.gray*0.001*z 
      }
      // update the anti-overlapping union
      #declare avoid_overlapping=union{
        object{avoid_overlapping}
        sphere{0,coc_radius scale <1,1,.01>*coc_spot_min
          scale (pow(coc_depth.gray,coc_pw)+.0000001)
          rotate coc_angle*z
          translate coc_pt*<width_ratio,height_ratio,1>
        }
      }
      #declare i_dens=i_dens+1;
    #end
  #end
#end

// print the value of the brightest spot found
#debug "\n"
#debug "max. hdr value: "
#debug str(max_val,5,3)
#debug "\n"
