// ***********************************
// camera effect : cross screen filter
// ***********************************
// Puts objects in front of the camera, just a bit over the viewing plane,
// tracing rays along a grid with adjustable density to evaluate the image 
// color at that point, using the color found to texturize the star objects

// a plane to trace into
#declare test_box=plane{z,1.1}

// pigment for evaluating the input image
#declare p_test=
pigment{
  input_image_map
  translate <-.5,-.5,0>
  scale 1+chromatic_aberration*aberration_amount*2 
}

// create the star object
#declare csf_object=
union{
  #declare i_points=0;
  #while (i_points<csf_points)
    cone{0,csf_radius,csf_length*y,.00001 rotate (360/csf_points)*i_points*z}
    #declare i_points=i_points+1; 
  #end
}

// in case of HDR, it will be useful to know wich was the value of the brightest spot,
// so we can adjust the trhesold value for the next render
#declare max_val=0;
// adjust the rows and cols to the image aspect ratio
#declare n_col=csf_density*width_ratio;
#declare n_row=csf_density*height_ratio;
// start placing csf objects on the grid
#declare i_col=0;
#while (i_col<n_col)
  #declare i_row=0;
  #while (i_row<n_row)
    // current point on the grid traced against the test box
    #declare csf_pt=<(-.5+i_col/n_col),(-.5+i_row/n_row),1>+<(1/n_col)*.25,(1/n_row)*.25,0>;
    #declare Norm=<0,0,0>;
    #declare Inter=trace(test_box,csf_pt,z,Norm);
    // evaluate the image color at this point
    #declare csf_int=eval_pigment(p_test,Inter);
    // find the value of the brightest spot on the image
    #if (csf_int.gray>max_val) 
      #declare max_val=csf_int.gray; 
    #end
    // place the csf objects on spots brighter than the desired threshold
    #if (csf_int.gray>csf_threshold)
      object{csf_object
        texture{
          pigment{
          spherical
          color_map{
            [0 rgbt 1]
            [1 rgbft csf_int+<0,0,0,0,1-csf_strength>]
          }
          scale csf_length*.5
          } 
          finish{emission 1}
        }
        // add layer of fake dispersion (will be transparent if csf_dispersion=0)
        texture{
          pigment{
          spherical frequency 1 poly_wave 1
          color_map{
            [0.00 rgbft csf_int+<0,0,0,0,1>]
            [0.20 rgbt <1,0,1,1-csf_dispersion>]
            [0.40 rgbt <0,0,1,1-csf_dispersion>]
            [0.60 rgbt <0,1,0,1-csf_dispersion>]
            [0.70 rgbt <1,1,0,1-csf_dispersion>]
            [0.80 rgbt <1,0,0,1-csf_dispersion>]
            [1.00 rgbft csf_int+<0,0,0,0,1-csf_dispersion>]
          }
          scale csf_length*.5
          } 
          finish{emission 1}
        }
        rotate csf_angle*z
        translate csf_pt*<width_ratio,height_ratio,1>
      }
    #end
    #declare i_row=i_row+1;
  #end
  #declare i_col=i_col+1;
#end

// print the value of the brightest spot found
#debug "\n"
#debug "max. hdr value: "
#debug str(max_val,5,3)
#debug "\n"
