// ******************************
// camera effect : luminous bloom
// ******************************
// Puts blobs in front of the camera, just a bit over the viewing plane,
// tracing rays along a grid with adjustable density to evaluate the image 
// color at that point, using the color found to texturize the blob object
// with a z gradient, so that the borders of the blob are transparent

// a plane to trace into
#declare test_box=plane{z,1.1}

// pigment for evaluating the input image
#declare p_test=
pigment{
  input_image_map
  translate <-.5,-.5,0>
  scale 1+chromatic_aberration*aberration_amount*2 
}

// to avoid excesive overlaping of the blob components, we will use an extra trace test 
// over a union of spheres representing previously placed components, so we first create
// something to add the test speheres later:
#declare avoid_overlapping=box{-.5,.5 translate 10*y} //just something away from the tracing area
// in case of HDR, it will be useful to know wich was the value of the brightest spot,
// so we can adjust the trhesold value for the next render
#declare max_val=0;
// start placing blob components on the grid
#declare n_col=bloom_density*width_ratio;
#declare n_row=bloom_density*height_ratio;
#declare i_col=0;
blob{threshold .5
#while (i_col<n_col)
  #declare i_row=0;
  #while (i_row<n_row)
    // current point on the grid traced against the test box
    #declare bloom_pt=<(-.5+i_col/n_col),(-.5+i_row/n_row),1>+<(1/n_col)*.25,(1/n_row)*.25,0>;
    #declare Norm=<0,0,0>;
    #declare Inter=trace(test_box,bloom_pt,z,Norm);
    // evaluate the image color at this point
    #declare bloom_int=eval_pigment(p_test,Inter);
    // find the value of the brightest spot on the image
    #if (bloom_int.gray>max_val) 
      #declare max_val=bloom_int.gray; 
    #end
    // place the blob components on spots brighter than the desired threshold
    #if (bloom_int.gray>bloom_brightness)
      #declare Norm=<0,0,0>;
      #declare Inter=trace(avoid_overlapping,bloom_pt,z,Norm);
      #if (Norm.gray=0)
        sphere{0,1,1 
          texture{
          pigment{
           gradient -z
           poly_wave bloom_pw
           color_map{
            [0 rgbft bloom_int+<0,0,0,0,1>]
            [1 rgbft bloom_int+<0,0,0,0,1-bloom_strength>]
           }
          } 
          finish{emission 1 diffuse 0 brilliance 0}
          }
          scale <bloom_amount*.1,bloom_amount*.1,bloom_amount*.1*.125>
          translate bloom_pt*<width_ratio,height_ratio,1>
        }
        // update the anti-overlapping union
        #declare avoid_overlapping=union{
          object{avoid_overlapping}
          sphere{0,bloom_amount*bloom_spot_min
          scale <bloom_amount*.1,bloom_amount*.1,bloom_amount*.1>
          translate bloom_pt*<width_ratio,height_ratio,1>
          }
        }
      #end
    #end
    #declare i_row=i_row+1;
  #end
  #declare i_col=i_col+1;
#end
}

// print the value of the brightest spot found
#debug "\n"
#debug "max. hdr value: "
#debug str(max_val,5,3)
#debug "\n"
