/*

  Jaime Vives Piqueres, Jul-2011
  --
  Post-Processing simulation of camera deffects (lens artifacts and aberrations)
  and some camera effects too...
  --
  to-do, but not even researched yet:
  + custom objects for coc (bokeh)
  + custom objects for csf (cross screen filter)
  + Random camera shake (as if the photographer was amateur, and forgot to enable anti-shaking)
  + Zoom burst (classical motion-blur like effect)
  + Digital sensor blow-out (suggested by Edouard Poor)
  + Corner softness (suggested by Edouard Poor)
  
*/

// --------------------------------------------------------------------------------- 
// PARAMETERS FOR THE EFFECTS: Turn them on/off and adjust their setup here

// **********************
// input image parameters
// **********************
// input image, remember to change the image type below
#declare image_file="examples/unable.jpg" 
#declare image_type=1;    // 0=hdr, 1=jpeg, 2=png, 3=tga
#declare image_gamma=2.2; // should correspond to the image assumed or embebbed gamma
#declare pw=0.95;          // polywave for image exposure control (<1=brighter, use if the image gets darkened by multiple effects)

// ********************
// chromatic aberration 
// ********************
#declare chromatic_aberration=1;  // on/off
#declare aberration_type=2;       // 1=transverse, 2=longitudinal, see http://toothwalker.org/optics/chromatic.html
#declare aberration_amount=.005;  // it should be much less for transverse than for longitudinal

// ****************
// shape distortion
// ****************
// + note: needs antialiasing as it bends existing lines
#declare optical_distortion=1;  // on/off
#declare distortion_amount=-.3; // positive for pincushion, negative for barrel

// ******************
// blue channel noise
// ******************
// + note: may need antialiasing, but not too much or will disapear
#declare blue_chanel_noise=0; // on/off
#declare bcn_amount=.5;       // 0-1, amount of noise

// **********
// film grain
// **********
// + note: may need antialiasing, but not too much or will disapear
#declare film_grain=1;           // on/off
#declare film_grain_amount=0.10;  // >0, amount of noise
#declare film_grain_midtone=0.3; // 0-1, 0=darkest tones, 1=brightest tones 

// **************
// luminous bloom
// **************
#declare luminous_bloom=0;    // on/off
#declare bloom_amount=.5;     // radius of the bloom blob components
#declare bloom_spot_min=.5;  // relative radius of the bright spots in the image
#declare bloom_brightness=.9; // minimum brightness level for the effect to show
#declare bloom_strength=.75;   // intensity of the effect
#declare bloom_density=200;    // denstity of the effect (grid rowsXcols)
#declare bloom_pw=4;          // poly_wave for fading control

// ******************
// cross creen filter 
// ******************
// + note: doesn't works with distortion at the moment
#declare cross_screen=0;      // on/off
#declare csf_points=6;        // number of star points (arms)
#declare csf_radius=.0025;    // radius of the star center 
#declare csf_length=.2;       // length of stars arms
#declare csf_threshold=.99;   // minimum brightness level for the effect to show 
#declare csf_strength=.25;    // intensity of the effect
#declare csf_angle=45;        // general rotation angle
#declare csf_density=50;      // density of the stars
#declare csf_dispersion=.1;   // amount of fake dispersion effect on the stars 

// *********************************
// coc (circles of confusion, bokeh)
// *********************************
// + note: doesn't works with distortion
#declare coc=0;             // on/off
#declare coc_radius=.05;    // radius of the confusion circles
#declare coc_spot_min=.2;   // relative radius of the bright spots in the image
#declare coc_threshold=.9;  // minimum brightness level for the effect to show 
#declare coc_strength=.75;  // intensity of the effect
#declare coc_blades=12;     // number of diaphragm blades: >=3!
#declare coc_angle=10;      // general rotation angle
#declare coc_density=5000;  // density of the stars
#declare coc_pw=.5;         // poly_wave for circle fading control
#declare r_coc=seed(387);   // radom seed for placing confusion circles
// Important! you need to supply a "focus map" below!! (see examples)
#declare focus_map="examples/reloj_focus-map.tga" 
#declare focus_map_type=3;    // 0=hdr, 1=jpeg, 2=png, 3=tga

// *************************
// dust on the lens (random)
// *************************
#declare dust_on_lens=1;   // on/off
#declare dust_amount=10;   // number of dust particles
#declare dust_size=1;      // size of the dust particles
#declare r_dust=seed(379); // radom seed for dust particles placement

// ******************
// optical vignetting
// ******************
#declare optical_vignetting=1;  // on/off
#declare vignetting_start=0.4;  // 0-1, distance from the center where the effect starts
#declare vignetting_end=0.7;    // 0-1, distance from the center where the effect is full
#declare vignetting_pw=2;       // polywave to control the effect ramp
#declare vignetting_filter=0.1; // 0=opaque to 1=transparent

// END OF EFFECTS PARAMETERS: No need to touch anything below this for normal usage
// --------------------------------------------------------------------------------

// basic common setup
#include "cdpp_common.inc"

// add thogther the components...
#if (chromatic_aberration)
  // ...with aberration if enabled...
  #include "cdpp_chromatic_aberration.inc"
#else
  // ...or untouched if disabled
  #declare t_output_image=
    texture{
      pigment{
        average
          pigment_map{
            [
              function{f_p(x,y,z).x/255}   
              poly_wave pw
              color_map{[0 rgb 0][1 Red*3]}
            ]
            [
              function{f_p(x,y,z).y/255} 
              poly_wave pw
              color_map{[0 rgb 0][1 Green*3]}
            ]
            [
              function{f_p(x,y,z).z/255} 
              poly_wave pw
              color_map{[0 rgb 0][1 Blue*3]}
            ]
          }
      }
      finish{emission 255 diffuse 0 brilliance 0}
    }
#end

// add blue chanel noise if enabled
#if (blue_chanel_noise) 
  #include "cdpp_blue_channel_noise.inc"
#end

// add film grain if enabled
#if (film_grain) 
  #include "cdpp_film_grain.inc"
#end

// add luminous bloom if enabled
#if (luminous_bloom)
  #include "cdpp_luminous_bloom.inc"
#end

// add cross-screen filter if enabled
#if (cross_screen)
  #include "cdpp_cross_screen_filter.inc"
#end

// add coc if enabled
#if (coc)
  #include "cdpp_coc.inc"
#end

// add dust on the lens if enabled
#if (dust_on_lens)
  #include "cdpp_dust_on_lens.inc"
#end

// add vignetting if enabled
#if (optical_vignetting)
  #include "cdpp_optical_vignetting.inc"
#end

// put the resulting texture on the screen plane
#include "screen.inc"
object{Screen_Plane(texture{t_output_image translate .5},1,<0,0,0>,<1,1,0>)
}
